(*
Windows向けunzip

unzipが古くてUTF-8に対応してないので、それっぽいパチモンを用意。
注意点
プログラムの起動時パラメタはAnsi（シフトJIS）。
Filename.dirname などはシフトJISでは「\」問題でおかしくなる。UTF-8を通せばOK。
Unix.mkdir や Zip.copy_entry_to_file は Ansiを通さないと文字化けする。

確認環境
OCaml 4.02.1 (WODI)
Windows 10
Cygwin

2022.05.30 Takeshi Yoneki
ostra@asahi.email.ne.jp
*)

type config = { 
  src: string ;
  dst: string ;
  listup: bool
}

let defaultConfig = { src = ""; dst = ""; listup = false }

let rec parseParam args cfg =
  match args with
  | p :: axs -> (
    match p with
    | "-d" -> (
      match args with
      | p :: v :: axs ->
        parseParam axs { cfg with dst = v }
      | _ ->
        parseParam axs cfg
    )
    | "-l" -> (
      parseParam axs { cfg with listup = true }
    )
    | _ -> (
      if cfg.src = "" then
        parseParam axs { cfg with src = p }
      else if cfg.dst = "" then
        parseParam axs { cfg with dst = p }
      else
        parseParam axs cfg
    )
  )
  | _ -> 
    cfg

let utf_8_flag = 0x800

let read_zip_general_purpose_bit_flag path =
  let ic = open_in_bin path in
  seek_in ic 6;
  let b0 = input_byte ic in
  let b1 = input_byte ic in
  close_in ic ;
  (b1 lsl 8) lor b0

let rec mkdir_all pathU8 =
  let pathU8 = Filename.concat (Filename.dirname pathU8) (Filename.basename pathU8) in
  let pathAnsi = TCharConv.utf8_to_ansi pathU8 in
  (*print_endline pathAnsi ;*)
  if Sys.file_exists pathAnsi then
    ()
  else (
    let parentU8 = Filename.dirname pathU8 in
    let parentAnsi = TCharConv.utf8_to_ansi parentU8 in
    if not (Sys.file_exists parentAnsi) then
      mkdir_all parentU8 ;
    Unix.mkdir pathAnsi 0775
  )

let proc cfg =
  let flag = read_zip_general_purpose_bit_flag cfg.src in
  let isUTF8 = (flag land utf_8_flag) <> 0 in
  let dstDirU8 = TCharConv.ansi_to_utf8 cfg.dst in
  let inz = Zip.open_in cfg.src in
  List.iter
    (fun e ->
      let fileNameU8 = if isUTF8 then e.Zip.filename else TCharConv.ansi_to_utf8 e.Zip.filename in
      let fileNameAnsi = TCharConv.utf8_to_ansi fileNameU8 in
      print_endline fileNameAnsi ;
      if not cfg.listup then (
        let pathU8 = Filename.concat dstDirU8 fileNameU8 in
        if e.Zip.is_directory then (
          mkdir_all pathU8
        )
        else (
          let dirU8 = Filename.dirname pathU8 in
          mkdir_all dirU8 ;
          let pathAnsi = TCharConv.utf8_to_ansi pathU8 in
          (*print_endline pathAnsi ;*)
          Zip.copy_entry_to_file inz e pathAnsi
        )
      )
    )  
    (Zip.entries inz) ;
  Zip.close_in inz
  
let main () =
  let cfg = parseParam (List.tl (Array.to_list Sys.argv)) defaultConfig in
  if cfg.src = "" then (
    print_endline "usage: unzip SRC_ZIP [-d DST_DIR -l]" ;
  )
  else (
    let cfg = 
      if cfg.dst = "" then
        { cfg with dst = "." }
      else
        cfg
    in
    proc cfg
  )

let _ = main ()
